"""
LogHandler.py
=============
"""
import logging
import os
import platform
from logging.config import dictConfig
from logging.handlers import RotatingFileHandler

import colors

from .Config import Configuration


class HostnameFilter(logging.Filter):
    hostname = platform.node()

    def filter(self, record):
        record.hostname = HostnameFilter.hostname
        return True


class HelperLogger(logging.Logger):
    """
    The HelperLogger is used by the application / gui as their logging class and *extends* the default python
    logger.logging class.

    This will separate the logging from the application / gui from that of the daemons.

    """

    runPath = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

    logPath = os.path.join(runPath, "log")

    if not os.path.exists(logPath):
        os.makedirs(logPath)

    config = Configuration()

    logDict = {
        "version": 1,
        "formatters": {
            "sysLogFormatter": {
                "format": "%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s"
            },
            "simpleFormatter": {
                "format": "%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s"
            },
        },
        "handlers": {
            "consoleHandler": {
                "class": "logging.StreamHandler",
                "level": "INFO",
                "stream": "ext://sys.stdout",
                "formatter": "simpleFormatter",
            }
        },
        "root": {"level": "DEBUG", "handlers": ["consoleHandler"]},
    }

    dictConfig(logDict)

    level_map = {
        "debug": "magenta",
        "info": "white",
        "warning": "yellow",
        "error": "red",
        "critical": "red",
    }

    def __init__(self, name, level=logging.NOTSET):

        super().__init__(name, level)

    def debug(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘DEBUG’ and color *MAGENTA.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        logger.debug(“Houston, we have a %s”, “thorny problem”, exc_info=1)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=HelperLogger.level_map["debug"])

        return super(HelperLogger, self).debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘INFO’ and color *WHITE*.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        logger.info(“Houston, we have a %s”, “interesting problem”, exc_info=1)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=HelperLogger.level_map["info"])

        return super(HelperLogger, self).info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘WARNING’ and color *YELLOW*.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        logger.warning(“Houston, we have a %s”, “bit of a problem”, exc_info=1)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=HelperLogger.level_map["warning"])

        return super(HelperLogger, self).warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘ERROR’ and color *RED*.

        Store logged message to the database for dashboard alerting.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        logger.error(“Houston, we have a %s”, “major problem”, exc_info=1)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=HelperLogger.level_map["error"])

        return super(HelperLogger, self).error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘CRITICAL’ and color *RED*.

        Store logged message to the database for dashboard alerting.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        logger.critical(“Houston, we have a %s”, “hell of a problem”, exc_info=1)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=HelperLogger.level_map["critical"])

        return super(HelperLogger, self).critical(msg, *args, **kwargs)


class UpdateHandler(HelperLogger):
    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)

        formatter = logging.Formatter(
            "%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s"
        )

        crf = RotatingFileHandler(
            filename=self.config.getUpdateLogFile(),
            maxBytes=self.config.getMaxLogSize(),
            backupCount=self.config.getBacklog(),
        )
        crf.setLevel(logging.DEBUG)
        crf.setFormatter(formatter)
        self.addHandler(crf)



import datetime
import hashlib
import json
import logging
import os
import shutil
from collections import namedtuple
from xml.sax import make_parser

import pymongo
from dateutil.parser import parse as parse_datetime
from pymongo import TEXT, ASCENDING

from .Config import Configuration
from .JSONFileHandler import JSONFileHandler
from .Toolkit import generate_title
from .XMLFileHandler import XMLFileHandler
from .content_handlers import CapecHandler, CWEHandler
from .db_action import DatabaseAction
from CveXplore.database.connection.mongo_db import MongoDBConnection

file_prefix = "nvdcve-1.1-"
file_suffix = ".json.gz"
file_mod = "modified"
file_rec = "recent"

date = datetime.datetime.now()
year = date.year + 1

# default config
defaultvalue = {"cwe": "Unknown"}

cveStartYear = Configuration.getCVEStartYear()


class CPEDownloads(JSONFileHandler):
    def __init__(self):
        self.feed_type = "CPE"
        self.prefix = "matches.item"
        super().__init__(self.feed_type, self.prefix)

        self.feed_url = Configuration.getFeedURL(self.feed_type.lower())

        self.logger = logging.getLogger("CPEDownloads")

    @staticmethod
    def process_cpe_item(item=None):
        if item is None:
            return None
        if "cpe23Uri" not in item:
            return None

        cpe = {
            "title": generate_title(item["cpe23Uri"]),
            "cpe_2_2": item["cpe23Uri"],
            "cpe_name": item["cpe_name"],
            "vendor": item["cpe23Uri"].split(":")[3],
            "product": item["cpe23Uri"].split(":")[4],
        }

        version_info = ""
        if "versionStartExcluding" in item:
            cpe["versionStartExcluding"] = item["versionStartExcluding"]
            version_info += cpe["versionStartExcluding"] + "_VSE"
        if "versionStartIncluding" in item:
            cpe["versionStartIncluding"] = item["versionStartIncluding"]
            version_info += cpe["versionStartIncluding"] + "_VSI"
        if "versionEndExcluding" in item:
            cpe["versionEndExcluding"] = item["versionEndExcluding"]
            version_info += cpe["versionEndExcluding"] + "_VEE"
        if "versionEndIncluding" in item:
            cpe["versionEndIncluding"] = item["versionEndIncluding"]
            version_info += cpe["versionEndIncluding"] + "_VEI"

        sha1_hash = hashlib.sha1(
            cpe["cpe_2_2"].encode("utf-8") + version_info.encode("utf-8")
        ).hexdigest()

        cpe["id"] = sha1_hash

        return cpe

    def process_item(self, item):
        cpe = self.process_cpe_item(item)

        if cpe is not None:
            if self.is_update:
                self.queue.put(
                    DatabaseAction(
                        action=DatabaseAction.actions.UpdateOne,
                        collection=self.feed_type.lower(),
                        doc=cpe,
                    )
                )
            else:
                self.queue.put(
                    DatabaseAction(
                        action=DatabaseAction.actions.InsertOne,
                        collection=self.feed_type.lower(),
                        doc=cpe,
                    )
                )

    def update(self, **kwargs):
        self.logger.info("CPE database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.is_update = False

        self.process_downloads([self.feed_url])

        self.logger.info("Finished CPE database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.logger.info("CPE Database population started")

        self.queue.clear()

        self.dropCollection(self.feed_type.lower())

        DatabaseIndexer().create_indexes(collection="cpe")

        self.is_update = False

        self.process_downloads([self.feed_url])

        self.logger.info("Finished CPE database population")

        return self.last_modified


class CVEDownloads(JSONFileHandler):
    def __init__(self):
        self.feed_type = "CVES"
        self.prefix = "CVE_Items.item"
        super().__init__(self.feed_type, self.prefix)

        self.feed_url = Configuration.getFeedURL("cve")
        self.modfile = file_prefix + file_mod + file_suffix
        self.recfile = file_prefix + file_rec + file_suffix

        self.logger = logging.getLogger("CVEDownloads")

    @staticmethod
    def get_cve_year_range():
        """
        Method to fetch the years where we need cve's for
        """
        for a_year in range(cveStartYear, year):
            yield a_year

    @staticmethod
    def get_cpe_info(cpeuri):
        query = {}
        version_info = ""
        if "versionStartExcluding" in cpeuri:
            query["versionStartExcluding"] = cpeuri["versionStartExcluding"]
            version_info += query["versionStartExcluding"] + "_VSE"
        if "versionStartIncluding" in cpeuri:
            query["versionStartIncluding"] = cpeuri["versionStartIncluding"]
            version_info += query["versionStartIncluding"] + "_VSI"
        if "versionEndExcluding" in cpeuri:
            query["versionEndExcluding"] = cpeuri["versionEndExcluding"]
            version_info += query["versionEndExcluding"] + "_VEE"
        if "versionEndIncluding" in cpeuri:
            query["versionEndIncluding"] = cpeuri["versionEndIncluding"]
            version_info += query["versionEndIncluding"] + "_VEI"

        return query, version_info

    @staticmethod
    def add_if_missing(cve, key, value):
        if value not in cve[key]:
            cve[key].append(value)
        return cve

    @staticmethod
    def get_vendor_product(cpeUri):
        vendor = cpeUri.split(":")[3]
        product = cpeUri.split(":")[4]
        return vendor, product

    @staticmethod
    def stem(cpeUri):
        cpeArr = cpeUri.split(":")
        return ":".join(cpeArr[:5])

    def process_cve_item(self, item=None):
        if item is None:
            return None
        if "ASSIGNER" not in item["cve"]["CVE_data_meta"]:
            item["cve"]["CVE_data_meta"]["ASSIGNER"] = None

        cve = {
            "id": item["cve"]["CVE_data_meta"]["ID"],
            "assigner": item["cve"]["CVE_data_meta"]["ASSIGNER"],
            "Published": parse_datetime(item["publishedDate"], ignoretz=True),
            "Modified": parse_datetime(item["lastModifiedDate"], ignoretz=True),
            "last-modified": parse_datetime(item["lastModifiedDate"], ignoretz=True),
        }

        for description in item["cve"]["description"]["description_data"]:
            if description["lang"] == "en":
                if "summary" in cve:
                    cve["summary"] += " {}".format(description["value"])
                else:
                    cve["summary"] = description["value"]
        if "impact" in item:
            cve["access"] = {}
            cve["impact"] = {}
            if "baseMetricV3" in item["impact"]:
                cve["impact3"] = {}
                cve["exploitability3"] = {}
                cve["impact3"]["availability"] = item["impact"]["baseMetricV3"][
                    "cvssV3"
                ]["availabilityImpact"]
                cve["impact3"]["confidentiality"] = item["impact"]["baseMetricV3"][
                    "cvssV3"
                ]["confidentialityImpact"]
                cve["impact3"]["integrity"] = item["impact"]["baseMetricV3"]["cvssV3"][
                    "integrityImpact"
                ]
                cve["exploitability3"]["attackvector"] = item["impact"]["baseMetricV3"][
                    "cvssV3"
                ]["attackVector"]
                cve["exploitability3"]["attackcomplexity"] = item["impact"][
                    "baseMetricV3"
                ]["cvssV3"]["attackComplexity"]
                cve["exploitability3"]["privilegesrequired"] = item["impact"][
                    "baseMetricV3"
                ]["cvssV3"]["privilegesRequired"]
                cve["exploitability3"]["userinteraction"] = item["impact"][
                    "baseMetricV3"
                ]["cvssV3"]["userInteraction"]
                cve["exploitability3"]["scope"] = item["impact"]["baseMetricV3"][
                    "cvssV3"
                ]["scope"]
                cve["cvss3"] = float(
                    item["impact"]["baseMetricV3"]["cvssV3"]["baseScore"]
                )
                cve["cvss3-vector"] = item["impact"]["baseMetricV3"]["cvssV3"][
                    "vectorString"
                ]
                cve["impactScore3"] = float(
                    item["impact"]["baseMetricV3"]["impactScore"]
                )
                cve["exploitabilityScore3"] = float(
                    item["impact"]["baseMetricV3"]["exploitabilityScore"]
                )
            else:
                cve["cvss3"] = None
            if "baseMetricV2" in item["impact"]:
                cve["access"]["authentication"] = item["impact"]["baseMetricV2"][
                    "cvssV2"
                ]["authentication"]
                cve["access"]["complexity"] = item["impact"]["baseMetricV2"]["cvssV2"][
                    "accessComplexity"
                ]
                cve["access"]["vector"] = item["impact"]["baseMetricV2"]["cvssV2"][
                    "accessVector"
                ]
                cve["impact"]["availability"] = item["impact"]["baseMetricV2"][
                    "cvssV2"
                ]["availabilityImpact"]
                cve["impact"]["confidentiality"] = item["impact"]["baseMetricV2"][
                    "cvssV2"
                ]["confidentialityImpact"]
                cve["impact"]["integrity"] = item["impact"]["baseMetricV2"]["cvssV2"][
                    "integrityImpact"
                ]
                cve["cvss"] = float(
                    item["impact"]["baseMetricV2"]["cvssV2"]["baseScore"]
                )
                cve["exploitabilityScore"] = float(
                    item["impact"]["baseMetricV2"]["exploitabilityScore"]
                )
                cve["impactScore"] = float(
                    item["impact"]["baseMetricV2"]["impactScore"]
                )
                cve["cvss-time"] = parse_datetime(
                    item["lastModifiedDate"], ignoretz=True
                )  # NVD JSON lacks the CVSS time which was present in the original XML format
                cve["cvss-vector"] = item["impact"]["baseMetricV2"]["cvssV2"][
                    "vectorString"
                ]
            else:
                cve["cvss"] = None
        if "references" in item["cve"]:
            cve["references"] = []
            for ref in item["cve"]["references"]["reference_data"]:
                cve["references"].append(ref["url"])
        if "configurations" in item:
            cve["vulnerable_configuration"] = []
            cve["vulnerable_product"] = []
            cve["vendors"] = []
            cve["products"] = []
            cve["vulnerable_product_stems"] = []
            cve["vulnerable_configuration_stems"] = []
            for cpe in item["configurations"]["nodes"]:
                if "cpe_match" in cpe:
                    for cpeuri in cpe["cpe_match"]:
                        if "cpe23Uri" not in cpeuri:
                            continue
                        if cpeuri["vulnerable"]:
                            query, version_info = self.get_cpe_info(cpeuri)
                            if query != {}:
                                query["id"] = hashlib.sha1(
                                    cpeuri["cpe23Uri"].encode("utf-8")
                                    + version_info.encode("utf-8")
                                ).hexdigest()
                                cpe_info = self.getCPEVersionInformation(query)
                                if cpe_info:
                                    if cpe_info["cpe_name"]:
                                        for vulnerable_version in cpe_info["cpe_name"]:
                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_product",
                                                vulnerable_version["cpe23Uri"],
                                            )
                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_configuration",
                                                vulnerable_version["cpe23Uri"],
                                            )
                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_configuration_stems",
                                                self.stem(
                                                    vulnerable_version["cpe23Uri"]
                                                ),
                                            )
                                            vendor, product = self.get_vendor_product(
                                                vulnerable_version["cpe23Uri"]
                                            )
                                            cve = self.add_if_missing(
                                                cve, "vendors", vendor
                                            )
                                            cve = self.add_if_missing(
                                                cve, "products", product
                                            )
                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_product_stems",
                                                self.stem(
                                                    vulnerable_version["cpe23Uri"]
                                                ),
                                            )
                                    else:
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_product",
                                            cpeuri["cpe23Uri"],
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_configuration",
                                            cpeuri["cpe23Uri"],
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_configuration_stems",
                                            self.stem(cpeuri["cpe23Uri"]),
                                        )
                                        vendor, product = self.get_vendor_product(
                                            cpeuri["cpe23Uri"]
                                        )
                                        cve = self.add_if_missing(
                                            cve, "vendors", vendor
                                        )
                                        cve = self.add_if_missing(
                                            cve, "products", product
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_product_stems",
                                            self.stem(cpeuri["cpe23Uri"]),
                                        )
                            else:
                                # If the cpe_match did not have any of the version start/end modifiers,
                                # add the CPE string as it is.
                                cve = self.add_if_missing(
                                    cve, "vulnerable_product", cpeuri["cpe23Uri"]
                                )
                                cve = self.add_if_missing(
                                    cve, "vulnerable_configuration", cpeuri["cpe23Uri"]
                                )
                                cve = self.add_if_missing(
                                    cve,
                                    "vulnerable_configuration_stems",
                                    self.stem(cpeuri["cpe23Uri"]),
                                )
                                vendor, product = self.get_vendor_product(
                                    cpeuri["cpe23Uri"]
                                )
                                cve = self.add_if_missing(cve, "vendors", vendor)
                                cve = self.add_if_missing(cve, "products", product)
                                cve = self.add_if_missing(
                                    cve,
                                    "vulnerable_product_stems",
                                    self.stem(cpeuri["cpe23Uri"]),
                                )
                        else:
                            cve = self.add_if_missing(
                                cve, "vulnerable_configuration", cpeuri["cpe23Uri"]
                            )
                            cve = self.add_if_missing(
                                cve,
                                "vulnerable_configuration_stems",
                                self.stem(cpeuri["cpe23Uri"]),
                            )
                if "children" in cpe:
                    for child in cpe["children"]:
                        if "cpe_match" in child:
                            for cpeuri in child["cpe_match"]:
                                if "cpe23Uri" not in cpeuri:
                                    continue
                                if cpeuri["vulnerable"]:
                                    query, version_info = self.get_cpe_info(cpeuri)
                                    if query != {}:
                                        query["id"] = hashlib.sha1(
                                            cpeuri["cpe23Uri"].encode("utf-8")
                                            + version_info.encode("utf-8")
                                        ).hexdigest()
                                        cpe_info = self.getCPEVersionInformation(query)
                                        if cpe_info:
                                            if cpe_info["cpe_name"]:
                                                for vulnerable_version in cpe_info[
                                                    "cpe_name"
                                                ]:
                                                    cve = self.add_if_missing(
                                                        cve,
                                                        "vulnerable_product",
                                                        vulnerable_version["cpe23Uri"],
                                                    )
                                                    cve = self.add_if_missing(
                                                        cve,
                                                        "vulnerable_configuration",
                                                        vulnerable_version["cpe23Uri"],
                                                    )
                                                    cve = self.add_if_missing(
                                                        cve,
                                                        "vulnerable_configuration_stems",
                                                        self.stem(
                                                            vulnerable_version[
                                                                "cpe23Uri"
                                                            ]
                                                        ),
                                                    )
                                                    (
                                                        vendor,
                                                        product,
                                                    ) = self.get_vendor_product(
                                                        vulnerable_version["cpe23Uri"]
                                                    )
                                                    cve = self.add_if_missing(
                                                        cve, "vendors", vendor
                                                    )
                                                    cve = self.add_if_missing(
                                                        cve, "products", product
                                                    )
                                                    cve = self.add_if_missing(
                                                        cve,
                                                        "vulnerable_product_stems",
                                                        self.stem(
                                                            vulnerable_version[
                                                                "cpe23Uri"
                                                            ]
                                                        ),
                                                    )
                                            else:
                                                cve = self.add_if_missing(
                                                    cve,
                                                    "vulnerable_product",
                                                    cpeuri["cpe23Uri"],
                                                )
                                                cve = self.add_if_missing(
                                                    cve,
                                                    "vulnerable_configuration",
                                                    cpeuri["cpe23Uri"],
                                                )
                                                cve = self.add_if_missing(
                                                    cve,
                                                    "vulnerable_configuration_stems",
                                                    self.stem(cpeuri["cpe23Uri"]),
                                                )
                                                (
                                                    vendor,
                                                    product,
                                                ) = self.get_vendor_product(
                                                    cpeuri["cpe23Uri"]
                                                )
                                                cve = self.add_if_missing(
                                                    cve, "vendors", vendor
                                                )
                                                cve = self.add_if_missing(
                                                    cve, "products", product
                                                )
                                                cve = self.add_if_missing(
                                                    cve,
                                                    "vulnerable_product_stems",
                                                    self.stem(cpeuri["cpe23Uri"]),
                                                )
                                    else:
                                        # If the cpe_match did not have any of the version start/end modifiers,
                                        # add the CPE string as it is.
                                        if "cpe23Uri" not in cpeuri:
                                            continue
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_product",
                                            cpeuri["cpe23Uri"],
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_configuration",
                                            cpeuri["cpe23Uri"],
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_configuration_stems",
                                            self.stem(cpeuri["cpe23Uri"]),
                                        )
                                        vendor, product = self.get_vendor_product(
                                            cpeuri["cpe23Uri"]
                                        )
                                        cve = self.add_if_missing(
                                            cve, "vendors", vendor
                                        )
                                        cve = self.add_if_missing(
                                            cve, "products", product
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_product_stems",
                                            self.stem(cpeuri["cpe23Uri"]),
                                        )
                                else:
                                    if "cpe23Uri" not in cpeuri:
                                        continue
                                    cve = self.add_if_missing(
                                        cve,
                                        "vulnerable_configuration",
                                        cpeuri["cpe23Uri"],
                                    )
                                    cve = self.add_if_missing(
                                        cve,
                                        "vulnerable_configuration_stems",
                                        self.stem(cpeuri["cpe23Uri"]),
                                    )
        if "problemtype" in item["cve"]:
            for problem in item["cve"]["problemtype"]["problemtype_data"]:
                for cwe in problem[
                    "description"
                ]:  # NVD JSON not clear if we can get more than one CWE per CVE (until we take the last one) -
                    # NVD-CWE-Other??? list?
                    if cwe["lang"] == "en":
                        cve["cwe"] = cwe["value"]
            if not ("cwe" in cve):
                cve["cwe"] = defaultvalue["cwe"]
        else:
            cve["cwe"] = defaultvalue["cwe"]
        cve["vulnerable_configuration_cpe_2_2"] = []
        return cve

    def process_item(self, item):
        cve = self.process_cve_item(item)

        if cve is not None:
            if self.is_update:
                self.queue.put(
                    DatabaseAction(
                        action=DatabaseAction.actions.UpdateOne,
                        collection=self.feed_type.lower(),
                        doc=cve,
                    )
                )
            else:
                self.queue.put(
                    DatabaseAction(
                        action=DatabaseAction.actions.InsertOne,
                        collection=self.feed_type.lower(),
                        doc=cve,
                    )
                )

    def update(self):
        self.logger.info("CVE database update started")

        # if collection is non-existent; assume it's not an update
        if "cves" not in self.getTableNames():
            self.is_update = False

        self.process_downloads(
            [self.feed_url + self.modfile, self.feed_url + self.recfile]
        )

        self.logger.info("Finished CVE database update")

        return self.last_modified

    def populate(self):
        urls = []

        self.logger.info("CVE database population started")

        self.logger.info(
            "Starting CVE database population starting from year: {}".format(
                cveStartYear
            )
        )

        self.is_update = False

        self.queue.clear()

        self.dropCollection("cves")

        DatabaseIndexer().create_indexes(collection="cves")

        for x in self.get_cve_year_range():
            getfile = file_prefix + str(x) + file_suffix

            urls.append(self.feed_url + getfile)

        self.process_downloads(urls)

        self.logger.info("Finished CVE database population")

        return self.last_modified


class VIADownloads(JSONFileHandler):
    def __init__(self):
        self.feed_type = "VIA4"
        self.prefix = "cves"
        super().__init__(self.feed_type, self.prefix)

        self.feed_url = Configuration.getFeedURL(self.feed_type.lower())

        self.logger = logging.getLogger("VIADownloads")

    def file_to_queue(self, file_tuple):

        working_dir, filename = file_tuple

        for cve in self.ijson_handler.fetch(filename=filename, prefix=self.prefix):
            x = 0
            for key, val in cve.items():
                entry_dict = {"id": key}
                entry_dict.update(val)
                self.process_item(item=entry_dict)
                x += 1

            self.logger.debug("Processed {} items from file: {}".format(x, filename))

        with open(filename, "rb") as input_file:
            data = json.loads(input_file.read().decode("utf-8"))

            self.setColInfo("via4", "sources", data["metadata"]["sources"])
            self.setColInfo("via4", "searchables", data["metadata"]["searchables"])

            self.logger.debug("Processed metadata from file: {}".format(filename))

        try:
            self.logger.debug("Removing working dir: {}".format(working_dir))
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(
                "Failed to remove working dir; error produced: {}".format(err)
            )

    def process_item(self, item):

        if self.is_update:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.UpdateOne,
                    collection=self.feed_type.lower(),
                    doc=item,
                )
            )
        else:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.InsertOne,
                    collection=self.feed_type.lower(),
                    doc=item,
                )
            )

    def update(self, **kwargs):
        self.logger.info("VIA4 database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.is_update = False

        self.process_downloads([self.feed_url])

        self.logger.info("Finished VIA4 database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.is_update = False
        self.queue.clear()
        return self.update()


class CAPECDownloads(XMLFileHandler):
    def __init__(self):
        self.feed_type = "CAPEC"
        super().__init__(self.feed_type)

        self.feed_url = Configuration.getFeedURL(self.feed_type.lower())

        self.logger = logging.getLogger("CAPECDownloads")

        # make parser
        self.parser = make_parser()
        self.ch = CapecHandler()
        self.parser.setContentHandler(self.ch)

    def file_to_queue(self, file_tuple):

        working_dir, filename = file_tuple

        self.parser.parse(filename)
        x = 0
        for attack in self.ch.capec:
            self.process_item(attack)
            x += 1

        self.logger.debug("Processed {} entries from file: {}".format(x, filename))

        try:
            self.logger.debug("Removing working dir: {}".format(working_dir))
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(
                "Failed to remove working dir; error produced: {}".format(err)
            )

    def update(self, **kwargs):
        self.logger.info("CAPEC database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.is_update = False

        self.process_downloads([self.feed_url])

        self.logger.info("Finished CAPEC database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.is_update = False
        self.queue.clear()
        return self.update()


class CWEDownloads(XMLFileHandler):
    def __init__(self):
        self.feed_type = "CWE"
        super().__init__(self.feed_type)

        self.feed_url = Configuration.getFeedURL(self.feed_type.lower())

        self.logger = logging.getLogger("CWEDownloads")

        # make parser
        self.parser = make_parser()
        self.ch = CWEHandler()
        self.parser.setContentHandler(self.ch)

    def file_to_queue(self, file_tuple):

        working_dir, filename = file_tuple

        self.parser.parse(filename)
        x = 0
        for cwe in self.ch.cwe:
            try:
                cwe["related_weaknesses"] = list(set(cwe["related_weaknesses"]))
            except KeyError:
                pass
            self.process_item(cwe)
            x += 1

        self.logger.debug("Processed {} entries from file: {}".format(x, filename))

        try:
            self.logger.debug("Removing working dir: {}".format(working_dir))
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(
                "Failed to remove working dir; error produced: {}".format(err)
            )

    def update(self, **kwargs):
        self.logger.info("CWE database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.is_update = False

        self.process_downloads([self.feed_url])

        self.logger.info("Finished CWE database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.is_update = False
        self.queue.clear()
        return self.update()


MongoUniqueIndex = namedtuple("MongoUniqueIndex", "index name unique weights")
MongoAddIndex = namedtuple("MongoAddIndex", "index name weights")


class DatabaseIndexer(object):
    def __init__(self):

        database = MongoDBConnection(**json.loads(os.getenv("MONGODB_CON_DETAILS")))
        self.database = database._dbclient

        self.indexes = {
            "cpe": [
                MongoUniqueIndex(
                    index=[("id", ASCENDING)],
                    name="id",
                    unique=True,
                    weights={"id": 10},
                ),
                MongoAddIndex(
                    index=[("vendor", ASCENDING)], name="vendor", weights={"vendor": 5}
                ),
                MongoAddIndex(
                    index=[("product", ASCENDING)],
                    name="product",
                    weights={"product": 3},
                ),
            ],
            "cpeother": [
                MongoUniqueIndex(
                    index=[("id", ASCENDING)],
                    name="id",
                    unique=True,
                    weights={"id": 10},
                )
            ],
            "cves": [
                MongoAddIndex(index=[("id", ASCENDING)], name="id", weights={"id": 10}),
                MongoAddIndex(
                    index=[("vulnerable_configuration", ASCENDING)],
                    name="vulnerable_configuration",
                    weights={"vulnerable_configuration": 3},
                ),
                MongoAddIndex(
                    index=[("vulnerable_product", ASCENDING)],
                    name="vulnerable_product",
                    weights={"vulnerable_product": 3},
                ),
                MongoAddIndex(
                    index=[("Modified", ASCENDING)],
                    name="Modified",
                    weights={"Modified": 3},
                ),
                MongoAddIndex(
                    index=[("Published", ASCENDING)],
                    name="Published",
                    weights={"Published": 3},
                ),
                MongoAddIndex(
                    index=[("last-modified", ASCENDING)],
                    name="last-modified",
                    weights={"last-modified": 3},
                ),
                MongoAddIndex(
                    index=[("cvss", ASCENDING)], name="cvss", weights={"cvss": 5}
                ),
                MongoAddIndex(
                    index=[("cvss3", ASCENDING)], name="cvss3", weights={"cvss3": 5}
                ),
                MongoAddIndex(
                    index=[("summary", TEXT)], name="summary", weights={"summary": 5}
                ),
                MongoAddIndex(
                    index=[("vendors", ASCENDING)],
                    name="vendors",
                    weights={"vendors": 5},
                ),
                MongoAddIndex(
                    index=[("products", ASCENDING)],
                    name="products",
                    weights={"products": 5},
                ),
                MongoAddIndex(
                    index=[("vulnerable_product_stems", ASCENDING)],
                    name="vulnerable_product_stems",
                    weights={"vulnerable_product_stems": 5},
                ),
                MongoAddIndex(
                    index=[("vulnerable_configuration_stems", ASCENDING)],
                    name="vulnerable_configuration_stems",
                    weights={"vulnerable_configuration_stems": 5},
                ),
            ],
            "via4": [
                MongoAddIndex(index=[("id", ASCENDING)], name="id", weights={"id": 10})
            ],
            "mgmt_whitelist": [
                MongoAddIndex(index=[("id", ASCENDING)], name="id", weights={"id": 10})
            ],
            "mgmt_blacklist": [
                MongoAddIndex(index=[("id", ASCENDING)], name="id", weights={"id": 10})
            ],
            "capec": [
                MongoAddIndex(
                    index=[("related_weakness", ASCENDING)],
                    name="related_weakness",
                    weights={"related_weakness": 10},
                )
            ],
        }

        self.logger = logging.getLogger("DatabaseIndexer")

    def getInfo(self, collection):
        return self.sanitize(self.database["info"].find_one({"db": collection}))

    def sanitize(self, x):
        if type(x) == pymongo.cursor.Cursor:
            x = list(x)
        if type(x) == list:
            for y in x:
                self.sanitize(y)
        if x and "_id" in x:
            x.pop("_id")
        return x

    def create_indexes(self, collection=None):

        if collection is not None:
            try:
                for each in self.indexes[collection]:
                    if isinstance(each, MongoUniqueIndex):
                        self.setIndex(
                            collection,
                            each.index,
                            name=each.name,
                            unique=each.unique,
                            weights=each.weights,
                        )
                    elif isinstance(each, MongoAddIndex):
                        self.setIndex(
                            collection, each.index, name=each.name, weights=each.weights
                        )
            except KeyError:
                # no specific index given, continue
                self.logger.warning(
                    "Could not find the requested collection: {}, skipping...".format(
                        collection
                    )
                )
                pass

        else:
            for index in self.iter_indexes():
                self.setIndex(index[0], index[1])

            for collection in self.indexes.keys():
                for each in self.indexes[collection]:
                    if isinstance(each, MongoUniqueIndex):
                        self.setIndex(
                            collection,
                            each.index,
                            name=each.name,
                            unique=each.unique,
                            weights=each.weights,
                        )
                    elif isinstance(each, MongoAddIndex):
                        self.setIndex(
                            collection, each.index, name=each.name, weights=each.weights
                        )

    def iter_indexes(self):
        for each in self.get_via4_indexes():
            yield each

    def get_via4_indexes(self):
        via4 = self.getInfo("via4")
        result = []
        if via4:
            for index in via4.get("searchables", []):
                result.append(("via4", index))
        return result

    def setIndex(self, col, field, **kwargs):
        try:
            self.database[col].create_index(field, **kwargs)
            self.logger.info("Success to create index %s on %s" % (field, col))
        except Exception as e:
            self.logger.error("Failed to create index %s on %s: %s" % (col, field, e))



"""
Cve Search API
==============
"""
import pymongo

from CveXplore.common.generic_api import GenericApi


class CveSearchApi(GenericApi):
    """
    The CveSearchApi handles the different arguments in order to perform a query to a specific Cve Search api endpoint.
    It mimics the pymongo cursor's behaviour to provide an ambiguous way to talk to either an API or a mongodb.
    """

    def __init__(self, db_collection, filter=None, limit=None, skip=None, sort=None):
        """
        Create a new CveSearchApi object.

        :param db_collection: ApiDatabaseCollection object
        :type db_collection: ApiDatabaseCollection
        :param filter: Filter to be used when querying data
        :type filter: dict
        :param limit: Limit value
        :type limit: int
        :param skip: Skip value
        :type skip: int
        :param sort: Sort value
        :type sort: tuple
        """
        super().__init__(
            address=tuple(db_collection.address),
            api_path=db_collection.api_path,
            proxies=db_collection.proxies,
            protocol=db_collection.protocol,
            user_agent=db_collection.user_agent,
        )
        from CveXplore.common.db_obj_mapping import database_objects_mapping

        self.database_objects_mapping = database_objects_mapping

        self.db_collection = db_collection

        self.collname = db_collection.collname

        data = filter
        if data is None:
            data = {}

        self.__data = {"retrieve": self.collname, "dict_filter": data}

        self.__empty = False

        self.__limit = limit
        self.__skip = skip
        self.__sort = sort

        self.data_queue = None

    def __repr__(self):
        """return a string representation of the obj GenericApi"""
        return "<<CveSearchApi:({}, {})>>".format(
            self.db_collection.address[0], self.db_collection.address[1]
        )

    def query(self):
        """
        Endpoint for free query to cve search data
        """

        results = self.call(method="POST", resource="/query", data=self.__data)

        if isinstance(results, str):
            self.data_queue = None
            return

        try:
            if len(results["data"]) > 0:
                self.data_queue = results["data"]
        except Exception:
            self.data_queue = results

    def limit(self, value):
        """
        Method to limit the amount of returned data

        :param value: Limit
        :type value: int
        :return: CveSearchApi object
        :rtype: CveSearchApi
        """

        if not isinstance(value, int):
            raise TypeError("limit must be an integer")

        self.__limit = value

        self.__data["limit"] = self.__limit

        return self

    def skip(self, value):
        """
        Method to skip the given amount of records before returning the data

        :param value: Skip
        :type value: int
        :return: CveSearchApi object
        :rtype: CveSearchApi
        """

        if not isinstance(value, int):
            raise TypeError("skip must be an integer")

        self.__skip = value

        self.__data["skip"] = self.__skip

        return self

    def sort(self, field, direction):
        """
        Method to sort the returned data

        :param field: Field to sort on
        :type field: str
        :param direction: The direction to sort in; e.g. pymongo.DESCENDING
        :type direction: int
        :return: CveSearchApi object
        :rtype: CveSearchApi
        """

        if direction == pymongo.DESCENDING:
            direction = "DESC"
        else:
            direction = "ASC"

        self.__sort = (field, direction)

        self.__data["sort"] = self.__sort[0]
        self.__data["sort_dir"] = self.__sort[1]

        return self

    def __iter__(self):
        """ Make this class an iterator """
        self.query()
        return self

    def next(self):
        """ Iterate to the results and return database objects """
        if self.__empty:
            raise StopIteration
        if self.data_queue is None:
            raise StopIteration
        try:
            if len(self.data_queue):
                return self.database_objects_mapping[self.collname](
                    **self.data_queue.pop()
                )
            else:
                raise StopIteration
        except TypeError:
            # We've received a Response object
            raise StopIteration

    __next__ = next



import shutil
from abc import abstractmethod

from .DownloadHandler import DownloadHandler
from .IJSONHandler import IJSONHandler


class JSONFileHandler(DownloadHandler):
    def __init__(self, feed_type, prefix):
        super().__init__(feed_type)

        self.is_update = True

        self.prefix = prefix

        self.ijson_handler = IJSONHandler()

    def __repr__(self):
        """ return string representation of object """
        return "<< JSONFileHandler:{} >>".format(self.feed_type)

    def file_to_queue(self, file_tuple):

        working_dir, filename = file_tuple

        # adjust the interval counter for debug logging when updating
        if self.is_update:
            interval = 500
        else:
            interval = 5000

        x = 0
        self.logger.debug("Starting processing of file: {}".format(filename))
        for cpe in self.ijson_handler.fetch(filename=filename, prefix=self.prefix):
            self.process_item(item=cpe)
            x += 1
            if x % interval == 0:
                self.logger.debug(
                    "Processed {} entries from file: {}".format(x, filename)
                )

        try:
            self.logger.debug("Removing working dir: {}".format(working_dir))
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(
                "Failed to remove working dir; error produced: {}".format(err)
            )

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def populate(self, **kwargs):
        raise NotImplementedError



"""
Main
====
"""
import functools
import json
import os
from collections import defaultdict

from pymongo import DESCENDING

from CveXplore.api.connection.api_db import ApiDatabaseSource
from CveXplore.common.db_mapping import database_mapping
from CveXplore.database.connection.mongo_db import MongoDBConnection
from CveXplore.errors import DatabaseIllegalCollection
from CveXplore.database.update.main_updater import MainUpdater

try:
    from version import VERSION
except ModuleNotFoundError:
    _PKG_DIR = os.path.dirname(__file__)
    version_file = os.path.join(_PKG_DIR, "VERSION")
    with open(version_file, "r") as fdsec:
        VERSION = fdsec.read()


class CveXplore(object):
    """
    Main class for CveXplore package
    """

    def __init__(self, mongodb_connection_details=None, api_connection_details=None):
        """
        Create a new instance of CveXplore

        :param mongodb_connection_details: Provide the connection details needed to establish a connection to a mongodb
                                           instance. The connection details should be in line with pymongo's
                                           documentation.
        :type mongodb_connection_details: dict
        :param api_connection_details: Provide the connection details needed to establish a connection to a cve-search
                                       API provider. The cve-search API provider should allow access to the 'query' POST
                                       endpoint; all other API endpoints are not needed for CveXplore to function. For
                                       the connection details supported please check the :ref:`API connection <api>`
                                       documentation.
        :type api_connection_details: dict
        """
        self.__version = VERSION

        if (
            api_connection_details is not None
            and mongodb_connection_details is not None
        ):
            raise ValueError(
                "CveXplore can be used to connect to either a cve-search database OR a cve-search api, not both!"
            )
        elif api_connection_details is None and mongodb_connection_details is None:
            # by default assume we are talking to a database
            mongodb_connection_details = {}
            os.environ["MONGODB_CON_DETAILS"] = json.dumps(mongodb_connection_details)
            self.datasource = MongoDBConnection(**mongodb_connection_details)
            self.database = MainUpdater(datasource=self.datasource)
        elif mongodb_connection_details is not None:
            os.environ["MONGODB_CON_DETAILS"] = json.dumps(mongodb_connection_details)
            self.datasource = MongoDBConnection(**mongodb_connection_details)
            self.database = MainUpdater(datasource=self.datasource)
        elif api_connection_details is not None:
            api_connection_details["user_agent"] = "CveXplore:{}".format(self.version)
            os.environ["API_CON_DETAILS"] = json.dumps(api_connection_details)
            self.datasource = ApiDatabaseSource(**api_connection_details)

        self.database_mapping = database_mapping

        from CveXplore.database.helpers.generic_db import GenericDatabaseFactory
        from CveXplore.database.helpers.specific_db import CvesDatabaseFunctions

        for each in self.database_mapping:
            if each != "cves":
                setattr(self, each, GenericDatabaseFactory(collection=each))
            else:
                setattr(self, each, CvesDatabaseFunctions(collection=each))

    def get_single_store_entry(self, entry_type, dict_filter={}):
        """
        Method to perform a query on a *single* collection in the data source and return a *single* result.

        :param entry_type: Which specific store are you querying? Choices are:
                           - capec;
                           - cpe;
                           - cwe;
                           - via4;
                           - cves;
        :type entry_type: str
        :param dict_filter: Dictionary representing a filter according to pymongo documentation
        :type dict_filter: dict
        :return: Objectified result from the query
        :rtype: object
        """

        entry_type = entry_type.lower()

        if entry_type not in self.database_mapping:
            raise DatabaseIllegalCollection(
                "Illegal collection requested: only {} are allowed!".format(
                    self.database_mapping
                )
            )

        result = getattr(self.datasource, "store_{}".format(entry_type)).find_one(
            dict_filter
        )

        return result

    def get_single_store_entries(self, query, limit=10):
        """
        Method to perform a query on a *single* collection in the data source and return all of the results.

        :param query: Tuple which contains the entry_type and the dict_filter in a tuple.
                      Choices for entry_type:
                      - capec;
                      - cpe;
                      - cwe;
                      - via4;
                      - cves;
                      dict_filter is a dictionary representing a filter according to pymongo documentation.
                      example:
                      get_single_store_entries(("CWE", {"id": "78"}))
        :type query: tuple
        :param limit: Limit the amount of returned results, defaults to 10
        :type limit: int
        :return: list with queried results
        :rtype: list
        """
        if not isinstance(query, tuple):
            raise ValueError(
                "Wrong parameter type, received: {} expected: tuple".format(type(query))
            )

        if len(query) != 2:
            raise ValueError(
                "Query parameter does not consist of the expected amount of variables, expected: 2 received: {}".format(
                    len(query)
                )
            )

        entry_type, dict_filter = query

        entry_type = entry_type.lower()

        if entry_type not in self.database_mapping:
            raise DatabaseIllegalCollection(
                "Illegal collection requested: only {} are allowed!".format(
                    self.database_mapping
                )
            )

        results = (
            getattr(self.datasource, "store_{}".format(entry_type))
            .find(dict_filter)
            .limit(limit)
        )

        return list(results)

    def get_multi_store_entries(self, *queries, limit=10):
        """
        Method to perform *multiple* queries on *a single* or *multiple* collections in the data source and return the
        results.

        :param queries: A list of tuples which contains the entry_type and the dict_filter.
                        Choices for entry_type:
                        - capec;
                        - cpe;
                        - cwe;
                        - via4;
                        - cves;
                        dict_filter is a dictionary representing a filter according to pymongo documentation.
                        example:
                        get_multi_store_entries([("CWE", {"id": "78"}), ("cves", {"id": "CVE-2009-0018"})])
        :type queries: list
        :return: Queried results in a single list
        :rtype: list
        """

        results = map(
            functools.partial(self.get_single_store_entries, limit=limit), *queries
        )

        joined_list = []

        for result_list in results:

            joined_list += result_list

        return list(joined_list)

    def cves_for_cpe(self, cpe_string, vuln_prod_search=False):
        """
        Method for retrieving Cves that match a single CPE string. By default the search will be made matching
        the configuration fields of the cves documents.

        :param cpe_string: CPE string: e.g. ``cpe:2.3:o:microsoft:windows_7:*:sp1:*:*:*:*:*:*``
        :type cpe_string: str
        :param vuln_prod_search: Search for matching products instead of configurations
        :type vuln_prod_search: bool
        :return: List with Cves
        :rtype: list
        """

        e = cpe_string.split(":")
        for x in range(0, 13 - len(e)):
            cpe_string += ":*"

        cpe = self.get_single_store_entry("cpe", {"cpe_2_2": cpe_string})

        if cpe is not None:
            return list(cpe.iter_cves_matching_cpe(vuln_prod_search))
        else:
            return cpe

    def cve_by_id(self, cve_id):
        """
        Method to retrieve a single CVE from the database by it's CVE ID number

        :param cve_id: String representing the CVE id; e.g. CVE-2020-0001
        :type cve_id: str
        :return: CVE object
        :rtype: Cves
        """
        return self.get_single_store_entry("cves", {"id": cve_id})

    def capec_by_cwe_id(self, cwe_id):
        """
        Method to retrieve capecs related to a specific CWE ID

        :param cwe_id: String representing the CWE id; e.g. '15'
        :type cwe_id:
        :return: List with Capecs
        :rtype: list
        """

        cwe = self.get_single_store_entry("cwe", {"id": cwe_id})

        if cwe is not None:
            return list(cwe.iter_related_capecs())
        else:
            return cwe

    def last_cves(self, limit=10):
        """
        Method to retrieve the last entered / changed cves. By default limited to 10.

        :return: List with Cves
        :rtype: list
        """

        results = (
            getattr(self.datasource, "store_cves")
            .find()
            .sort("Modified", DESCENDING)
            .limit(limit)
        )

        return list(results)

    def get_db_content_stats(self):
        """
        Property returning the stats from the database. Stats consist of the time last modified and the document count
        per cvedb store in the database.

        :return: Statistics
        :rtype: dict
        """

        stats = defaultdict(dict)

        if not isinstance(self.datasource, ApiDatabaseSource):

            if hasattr(self.datasource, "store_info"):
                results = self.datasource.store_info.find({})
                for each in results:
                    each.pop("_id")
                    db = each["db"]
                    each.pop("db")

                    if "sources" in each:
                        each.pop("sources")
                        each.pop("searchables")

                    each["last-modified"] = str(each["last-modified"])
                    each["document count"] = getattr(
                        self.datasource, "store_{}".format(db)
                    ).count()
                    stats[db] = each

                    for mgmtlist in ["mgmt_blacklist", "mgmt_whitelist"]:
                        stats[mgmtlist] = {
                            "document count": getattr(
                                self.datasource, "store_{}".format(mgmtlist)
                            ).count()
                        }

                return dict(stats)

            else:
                return "Database info could not be retrieved"

        return "Using api endpoint: {}".format(self.datasource.url)

    @property
    def version(self):
        """ Property returning current version """
        return self.__version

    def __repr__(self):
        """ String representation of object """
        return "<< CveXplore:{} >>".format(self.version)



import datetime
import os
import re

runPath = os.path.dirname(os.path.realpath(__file__))


class Configuration(object):
    CVE_START_YEAR = os.getenv("CVE_START_YEAR", 2002)

    SOURCES = os.getenv(
        "SOURCES",
        {
            "cve": "https://nvd.nist.gov/feeds/json/cve/1.1/",
            "cpe": "https://nvd.nist.gov/feeds/json/cpematch/1.0/nvdcpematch-1.0.json.zip",
            "cwe": "https://cwe.mitre.org/data/xml/cwec_v4.4.xml.zip",
            "capec": "https://capec.mitre.org/data/xml/capec_v3.4.xml",
            "via4": "https://www.cve-search.org/feeds/via4.json",
        },
    )

    HTTP_PROXY = os.getenv("HTTP_PROXY", "")

    LOGGING_MAX_FILE_SIZE = os.getenv("LOGGING_MAX_FILE_SIZE", "100MB")
    LOGGING_BACKLOG = os.getenv("LOGGING_BACKLOG", 5)
    LOGGING_FILE_NAME = os.getenv("LOGGING_FILE_NAME", "./log/update_populate.log")

    @classmethod
    def getCVEStartYear(cls):
        next_year = datetime.datetime.now().year + 1
        start_year = cls.CVE_START_YEAR
        if start_year < cls.CVE_START_YEAR or start_year > next_year:
            print(
                "The year %i is not a valid year.\ndefault year %i will be used."
                % (start_year, cls.default["CVEStartYear"])
            )
            start_year = cls.default["CVEStartYear"]
        return start_year

    @classmethod
    def getProxy(cls):
        return cls.HTTP_PROXY

    @classmethod
    def getFeedURL(cls, source):
        return cls.SOURCES[source]

    @classmethod
    def toPath(cls, path):
        return path if os.path.isabs(path) else os.path.join(runPath, "../..", path)

    @classmethod
    def getUpdateLogFile(cls):
        return cls.toPath(cls.LOGGING_FILE_NAME)

    @classmethod
    def getMaxLogSize(cls):
        size = cls.LOGGING_MAX_FILE_SIZE
        split = re.findall("\d+|\D+", size)
        multipliers = {"KB": 1024, "MB": 1024 * 1024, "GB": 1024 * 1024 * 1024}
        if len(split) == 2:
            base = int(split[0])
            unit = split[1].strip().upper()
            return base * multipliers.get(unit, 1024 * 1024)
        # if size is not a correctly defined set it to 100MB
        else:
            return 100 * 1024 * 1024

    @classmethod
    def getBacklog(cls):
        return cls.LOGGING_BACKLOG



import click

from CveXplore.cli_cmds.cli_utils.utils import printer


@click.group("stats", invoke_without_command=True)
@click.option("-d", "--datasource", is_flag=True, help="Show the datasource statistics")
@click.option("--pretty", is_flag=True, help="Pretty print the output")
@click.pass_context
def stats_cmd(ctx, datasource, pretty, output="json"):
    if ctx.invoked_subcommand is None:
        printer(
            input_data=ctx.obj["data_source"].get_db_content_stats,
            pretty=pretty,
            output=output,
        )
    else:
        printer(
            input_data=ctx.obj["data_source"].get_db_content_stats,
            pretty=pretty,
            output=output,
        )



class DatabaseException(Exception):
    pass


class DatabaseEmptyException(DatabaseException):
    pass


class DatabaseConnectionException(DatabaseException):
    pass


class DatabaseIllegalCollection(DatabaseException):
    pass



"""
capec
=====
"""
from CveXplore.common.data_source_connection import DatasourceConnection


class Capec(DatasourceConnection):
    """
    Capec database object
    """

    def __init__(self, **kwargs):

        super().__init__("capec")

        for each in kwargs:
            setattr(self, each, kwargs[each])

    def iter_related_weaknessess(self):
        """
        Generator function for iterating the related weaknesses from the current Capec object

        :return: Cwe object
        :rtype: Cwe
        """
        if hasattr(self, "related_weakness"):
            if len(self.related_weakness) != 0:
                for each in self.related_weakness:

                    cwe_doc = self._datasource_connection.store_cwe.find_one(
                        {"id": each}
                    )

                    yield cwe_doc

    def iter_related_capecs(self):
        """
        Generator function for iterating the related capecs from the current Capec object

        :return: Capec objects
        :rtype: Capec
        """
        if hasattr(self, "related_capecs"):
            if len(self.related_capecs) != 0:
                for each in self.related_capecs:
                    capec_doc = self._datasource_connection.store_capec.find_one(
                        {"id": each}
                    )

                    yield capec_doc

    def to_dict(self):
        """
        Method to convert the entire object to a dictionary

        :return: Data from object
        :rtype: dict
        """

        return {k: v for (k, v) in self.__dict__.items() if not k.startswith("_")}

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__

    def __repr__(self):
        """ String representation of object """
        return "<< Capec:{} >>".format(self.id)



from abc import abstractmethod

from .DownloadHandler import DownloadHandler
from .db_action import DatabaseAction


class XMLFileHandler(DownloadHandler):
    def __init__(self, feed_type):
        super().__init__(feed_type)
        self.is_update = True

    def __repr__(self):
        """ return string representation of object """
        return "<< XMLFileHandler:{} >>".format(self.feed_type)

    def process_item(self, item):

        if self.is_update:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.UpdateOne,
                    collection=self.feed_type.lower(),
                    doc=item,
                )
            )
        else:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.InsertOne,
                    collection=self.feed_type.lower(),
                    doc=item,
                )
            )

    @abstractmethod
    def file_to_queue(self, *args):
        raise NotImplementedError

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def populate(self, **kwargs):
        raise NotImplementedError



"""
Mongo DB connection
===================
"""
import atexit

from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError

from CveXplore.database.helpers.cvesearch_mongo_database import CveSearchCollection
from CveXplore.errors import DatabaseConnectionException


class MongoDBConnection(object):
    """
    The MongoDBConnection class serves as a shell that functions as uniform way to connect to the mongodb backend.
    By default it will try to establish a connection towards a mongodb running on localhost (default port 27017) and
    database 'cvedb' (as per defaults of cve_search)
    """

    def __init__(
        self, host="mongodb://127.0.0.1:27017", port=None, database="cvedb", **kwargs
    ):
        """


        :param host: The `host` parameter can be a full `mongodb URI
                     <http://dochub.mongodb.org/core/connections>`_, in addition to
                     a simple hostname.
        :type host: str
        :param port: Port number (optional when a URI is used as host parameter)
        :type port: int
        :param database: Database to connect to; defaults to cvedb (Cve Search default)
        :type database: str
        :param kwargs: Other arguments supported by the MongoClient instantiation
        :type kwargs: dict
        """

        self.client = None
        self._dbclient = None

        self.client = MongoClient(host, port, connect=False, **kwargs)

        self._dbclient = self.client[database]

        try:
            collections = self._dbclient.list_collection_names()
        except ServerSelectionTimeoutError as err:
            raise DatabaseConnectionException(
                "Connection to the database failed: {}".format(err)
            )

        if len(collections) != 0:
            for each in collections:
                self.__setattr__(
                    "store_{}".format(each),
                    CveSearchCollection(database=self._dbclient, name=each),
                )

        atexit.register(self.disconnect)

    def get_collections_details(self):

        for each in self._dbclient.list_collections():
            yield each

    def set_handlers_for_collections(self):
        for each in self._dbclient.list_collection_names():
            if not hasattr(self, each):
                setattr(
                    self,
                    "store_{}".format(each),
                    CveSearchCollection(database=self._dbclient, name=each),
                )

    @property
    def get_collection_names(self):

        return self._dbclient.list_collection_names()

    def disconnect(self):
        """
        Disconnect from mongodb
        """
        if self.client is not None:
            self.client.close()

    def __del__(self):
        """Called when the class is garbage collected."""
        self.disconnect()

    def __repr__(self):
        """ String representation of object """
        return "<< MongoDBConnection:{} >>".format(self._dbclient.name)



import logging

import ijson

from .LogHandler import UpdateHandler

logging.setLoggerClass(UpdateHandler)


class IJSONHandler(object):
    def __init__(self):
        self.logger = logging.getLogger("IJSONHandler")

    def fetch(self, filename, prefix):
        x = 0
        with open(filename, "rb") as input_file:
            for item in ijson.items(input_file, prefix):
                yield item
                x += 1

        self.logger.debug(
            "Processed {} items from file: {}, using prefix: {}".format(
                x, filename, prefix
            )
        )



"""
generic_api.py
==============
"""
import json
from json import JSONDecodeError

import requests


class GenericApi(object):
    """
    The GenericApi class serves as a base class for all API's used within CveXplore
    """

    def __init__(
        self,
        address,
        api_path=None,
        proxies=None,
        protocol="https",
        user_agent="GenericApi",
    ):
        """
        The Generic api caller handles all communication towards a api resource.

        :param address: Tuple with host ip/name and port
        :type address: tuple
        :param api_path: Generic to connect to api resources, defaults to 'None'
        :type api_path: str
        :param proxies: If you need to use a proxy, you can configure individual requests with the proxies argument
                        to any request method
        :type proxies: dict
        :param protocol: Protocol to use when connecting to api; defaults to 'https'
        :type protocol: str
        """

        if not isinstance(address, tuple):
            raise TypeError(
                "The parameter 'address' has to be a tuple with the address and port of the api host"
                "e.g. ('127.0.0.1', 8834) "
            )

        self.verify = False
        self.server = address[0]
        self.port = address[1]
        self.protocol = protocol
        self.baseurl = "{}://{}:{}".format(self.protocol, self.server, self.port)
        self.api_path = api_path
        self.proxies = proxies
        self.user_agent = user_agent

        self.myheaders = self.__default_headers

    def __repr__(self):
        """return a string representation of the obj GenericApi"""
        return "<<GenericApi:({}, {})>>".format(self.server, self.port)

    def __build_url(self, resource):
        """
        Internal method to build a url to use when executing commands

        :param resource: API end point to connect to
        :type resource: str
        :return: url string
        :rtype: str
        """
        if self.api_path is None:
            return "{0}/{1}".format(self.baseurl, resource)
        else:
            return "{0}/{1}/{2}".format(self.baseurl, self.api_path, resource)

    def __connect(self, method, resource, session, data=None, timeout=60):
        """
        Send a request

        Send a request to api host based on the specified data. Specify the content type as JSON and
        convert the data to JSON format.

        :param method: http method to use (e.g. POST, GET, DELETE, PUT)
        :type method: str
        :param resource: API end point to connect to
        :type resource: str
        :param session: Session object from requests library
        :type session: object
        :param data: Request body data
        :type data: dict
        :param timeout: Set the timeout on a request, defaults to 60 seconds
        :type timeout: int
        :return: response from the server
        :rtype: dict
        """

        requests.packages.urllib3.disable_warnings()

        if data is None:
            request_api_resource = {
                "headers": self.myheaders,
                "verify": self.verify,
                "timeout": timeout,
                "proxies": self.proxies,
            }
        else:
            data = json.dumps(data)
            request_api_resource = {
                "data": data,
                "headers": self.myheaders,
                "verify": self.verify,
                "timeout": timeout,
                "proxies": self.proxies,
            }

        if method == "POST":
            r = session.post(self.__build_url(resource), **request_api_resource)
        elif method == "PUT":
            r = session.put(self.__build_url(resource), **request_api_resource)
        elif method == "DELETE":
            r = session.delete(self.__build_url(resource), **request_api_resource)
        else:
            r = session.get(self.__build_url(resource), **request_api_resource)

        try:
            json_response = json.loads(r.text)
        except JSONDecodeError:
            json_response = r

        return json_response

    def call(self, method=None, resource=None, data=None):
        """
        Method for requesting free format api resources

        :param method: http method to use (e.g. POST, GET, DELETE, PUT)
        :type method: str
        :param resource: API end point to connect to
        :type resource: str
        :param data: Request body data
        :type data: dict
        :return: query result
        :rtype: dict
        """
        try:
            with requests.Session() as session:
                result = self.__connect(
                    method=method, resource=resource, session=session, data=data
                )
                return result
        except requests.ConnectionError:
            print("Connection error, is the host up?")
            raise

    @property
    def __default_headers(self):
        """
        Property to return the default headers

        :return: Default headers
        :rtype: dict
        """
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": "{}".format(self.user_agent),
        }

    @property
    def headers(self):
        """
        Property to return the current headers

        :return: self.myheaders
        :rtype: dict
        """

        return self.myheaders

    def set_header_field(self, field, value):
        """
        Method to add a header and set it's value

        :param field: Name of the header field to add
        :type field: str
        :param value: Value of the header field
        :type value: str
        :return: self.myheaders
        :rtype: dict
        """

        self.myheaders[field] = value

        return self.myheaders

    def del_header_field(self, field):
        """
        Method to delete a header field

        :param field: Name of the header field to delete
        :type field: str
        :return: self.myheaders
        :rtype: dict
        """

        self.myheaders.pop(field)

        return self.myheaders

    def reset_headers(self):
        """
        Method to reset the headers to the default values
        """

        self.myheaders = self.__default_headers



import click


class Mutex(click.Option):
    def __init__(self, *args, **kwargs):
        self.not_required_if: list = kwargs.pop("not_required_if")

        if not hasattr(self, "not_required_if"):
            raise AssertionError("'not_required_if' parameter required")

        kwargs["help"] = (
            kwargs.get("help", "")
            + ". *** Option is mutually exclusive with the option(s): '"
            + ", ".join(self.not_required_if)
            + "'. ***"
        ).strip()
        super(Mutex, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        current_opt: bool = self.name in opts
        for mutex_opt in self.not_required_if:
            if mutex_opt in opts:
                if current_opt:
                    raise click.UsageError(
                        "Illegal usage: '--"
                        + str(self.name)
                        + "' is mutually exclusive with the option(s): '--"
                        + str(mutex_opt)
                        + "'."
                    )
                else:
                    self.prompt = None
        return super(Mutex, self).handle_parse_result(ctx, opts, args)



import os

from setuptools import setup, find_packages

from version import VERSION

# The directory containing this file
HERE = os.path.abspath(os.path.dirname(__file__))

# The text of the README file
with open(os.path.join(HERE, "README.rst")) as fid:
    README = fid.read().split("##INCLUDE_MARKER##")[1]

with open(os.path.join(HERE, "requirements.txt")) as fid:
    REQS = fid.read().splitlines()

setup(
    name="CveXplore",
    version=VERSION,
    packages=find_packages(exclude=("tests", "docs")),
    url="https://github.com/cve-search/CveXplore",
    license="GNU General Public License v3.0",
    author="Paul Tikken",
    author_email="paul.tikken@gmail.com",
    home_page="https://github.com/cve-search/CveXplore",
    project_urls={
        "Documentation": "https://cve-search.github.io/CveXplore/",
        "Issues": "https://github.com/cve-search/CveXplore/issues",
    },
    description="Package for interacting with cve-search",
    long_description=README,
    long_description_content_type="text/x-rst",
    package_data={"CveXplore": ["LICENSE", "VERSION"]},
    include_package_data=True,
    entry_points='''
            [console_scripts]
            cvexplore=CveXplore.cli:main
        ''',
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=REQS,
)



import click


@click.group(
    "database", invoke_without_command=True, help="Database update / populate commands"
)
@click.pass_context
def db_cmd(ctx):
    pass


@db_cmd.group("update", invoke_without_command=True, help="Update the database")
def update_cmd(ctx):
    ctx.obj["data_source"].database.update()


@db_cmd.group("populate", invoke_without_command=True, help="Populate the database")
def populate_cmd(ctx):
    ctx.obj["data_source"].database.populate()



from CveXplore.database.update.Sources_process import (
    CPEDownloads,
    CVEDownloads,
    CWEDownloads,
    CAPECDownloads,
    VIADownloads,
    DatabaseIndexer,
)


class MainUpdater(object):
    def __init__(self, datasource, repopulate=False):

        self.datasource = datasource

        self.repopulate = repopulate

        self.sources = [
            {"name": "cpe", "updater": CPEDownloads},
            {"name": "cve", "updater": CVEDownloads},
            {"name": "cwe", "updater": CWEDownloads},
            {"name": "capec", "updater": CAPECDownloads},
            {"name": "via4", "updater": VIADownloads},
        ]

        self.posts = [{"name": "ensureindex", "updater": DatabaseIndexer}]

    def update(self):

        for source in self.sources:
            up = source["updater"]()
            up.update()

        for post in self.posts:
            indexer = post["updater"]()
            indexer.create_indexes()

        self.datasource.set_handlers_for_collections()

    def populate(self):

        for source in self.sources:
            populator = source["updater"]()
            populator.populate()

        for post in self.posts:
            indexer = post["updater"]()
            indexer.create_indexes()

        self.datasource.set_handlers_for_collections()



"""
API connection
==============
"""

from CveXplore.api.helpers.cve_search_api import CveSearchApi
from CveXplore.common.db_mapping import database_mapping


class ApiDatabaseSource(object):
    """
    The ApiDatabaseSource mimics the behaviour of the MongoDBConnection.
    """

    def __init__(
        self,
        address,
        api_path=None,
        proxies=None,
        protocol="https",
        user_agent="CveXplore",
    ):
        """
        Create new instance of the ApiDatabaseSource

        :param address: Tuple with host ip/name and port
        :type address: tuple
        :param api_path: The api_path parameter needs to be provided if the API runs on a non-root path.
                         So if the API is reachable at: https://localhost/api, this parameter needs to be set to 'api'.
                         So it is needed to connect to api resources, defaults to 'None'
        :type api_path: str
        :param proxies: If you need to use a proxy, you can configure individual requests with the proxies argument
                        to any request method
        :type proxies: dict
        :param protocol: Protocol to use when connecting to api; defaults to 'https'
        :type protocol: str
        :param user_agent: User agent to use when connecting; defaults to CveXplore:<<version>>
        :type user_agent: str
        """

        self.database_mapping = database_mapping

        if address == {}:
            return

        self.address = address

        if api_path is not None:
            self.url = "{}://{}:{}/{}".format(
                protocol, address[0], address[1], api_path
            )
        else:
            self.url = "{}://{}:{}".format(protocol, address[0], address[1])

        for each in self.database_mapping:
            setattr(
                self,
                "store_{}".format(each),
                ApiDatabaseCollection(
                    address=address,
                    api_path=api_path,
                    proxies=proxies,
                    protocol=protocol,
                    user_agent=user_agent,
                    collname=each,
                ),
            )

    def __repr__(self):
        """return a string representation of the obj ApiDatabaseSource"""
        return "<< ApiDatabaseSource: {} >>".format(self.address)


class ApiDatabaseCollection(object):
    """
    The ApiDatabaseCollection mimics the behaviour of the CveSearchCollection
    """

    def __init__(
        self,
        address,
        collname,
        api_path=None,
        proxies=None,
        protocol="https",
        user_agent="CveXplore",
    ):
        """
        Create a new ApiDatabaseCollection.

        :param address: Tuple with host ip/name and port
        :type address: tuple
        :param collname: Collection name
        :type collname: str
        :param api_path: The api_path parameter needs to be provided if the API runs on a non-root path.
                         So if the API is reachable at: https://localhost/api, this parameter needs to be set to 'api'.
                         So it is needed to connect to api resources, defaults to 'None'
        :type api_path: str
        :param proxies: If you need to use a proxy, you can configure individual requests with the proxies argument
                        to any request method
        :type proxies: dict
        :param protocol: Protocol to use when connecting to api; defaults to 'https'
        :type protocol: str
        :param user_agent: User agent to use when connecting; defaults to CveXplore:<<version>>
        :type user_agent: str
        """

        self.address = address
        self.api_path = api_path
        self.proxies = proxies
        self.protocol = protocol
        self.user_agent = user_agent

        self.collname = collname

    def find(self, filter=None):
        """
        Query the api endpoint as you would do so with a pymongo Collection.

        :return: Reference to the CveSearchApi
        :rtype: CveSearchApi
        """

        return CveSearchApi(self, filter)

    def find_one(self, filter=None):
        """
        Query the api endpoint as you would do so with a pymongo Collection.

        :return: Data or None
        :rtype: object
        """

        cursor = self.find(filter)

        for result in cursor.limit(-1):
            return result
        return None

    def __repr__(self):
        """return a string representation of the obj ApiDatabaseCollection"""
        return "<< ApiDatabaseCollection: {} >>".format(self.address)
